import logging
import sys
import os.path

from yaml import load as yaml_load

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from .common import singlton_class

console_logger = logging.getLogger('console_logger')


@singlton_class
class Config():

    def __init__(self, filename) -> None:
        self.filename = filename
        try:
            with open(filename, 'r') as f:
                self.conf_dict = yaml_load(f, Loader)
        except FileNotFoundError:
            console_logger.error('Config file not found. Exiting.')
            sys.exit(1)

        for fname in (
            'inventory_file',
            'keepass_db_file',
            'keepass_key_file'
        ):
            if fname not in self.conf_dict.keys():
                console_logger.error(
                    '{} is not configured. Exiting.'.format(fname)
                )
                sys.exit(1)

        for k, v in self.conf_dict.items():
            setattr(self, k, v)

        for fname in (
            self.inventory_file,
            self.keepass_db_file,
            self.keepass_key_file
        ):
            if not os.path.exists(fname):
                console_logger.error(
                    'File {} not found. Exiting'.format(fname)
                )
                sys.exit(1)

    def get_node_config(self, node_name) -> dict:
        pass



import logging
import os
import sys
from typing import List

from yaml import load as yaml_load
try:
    from yaml import CLoader as YamlLoader
except ImportError:
    from yaml import Loader as YamlLoader

logger = logging.getLogger(__name__)
console_logger = logging.getLogger('console_logger')

HOST_REQUIRED_FIELDS = {'hostname', 'credentials'}


class Inventory():

    def __init__(self, inventory_file) -> None:
        # inventory_path = os.path.abspath(inventory_file)
        inventory_path = os.path.abspath(os.path.expanduser(inventory_file))
        # yaml.
        try:
            with open(inventory_path, 'r') as f:
                self.inventory_dict = yaml_load(f, YamlLoader)
        except FileNotFoundError:
            console_logger.error('Inventory file not found. Exiting')
            sys.exit(1)

        self._flat = self._flatten_v2()

    def __len__(self):
        return len(self._flat)

    def __contains__(self, name):
        if name in self._flat.keys():
            return True
        return False

    def __iter__(self):
        return iter(self._flat)

    def __getitem__(self, hostname):
        host_config = self._flat.get(hostname, None)
        console_logger.debug('Got "{}" config: {}'.format(
            hostname,
            host_config)
        )
        if host_config:
            if (HOST_REQUIRED_FIELDS & set(host_config.keys()) !=
                    HOST_REQUIRED_FIELDS):
                missing_fields = (HOST_REQUIRED_FIELDS - (
                    HOST_REQUIRED_FIELDS & set(host_config.keys())
                ))
                console_logger.error(
                    'Host "{}" config is missing required fields {}'.format(
                        hostname, missing_fields
                    )
                )
                host_config = None
        else:
            console_logger.error(
                'Host "{}" not found!'.format(hostname)
            )
        return host_config

    def _flatten(self) -> List:
        flattened_dict = []

        def do_flat(d, prefix=''):
            new_prefix = prefix
            for k in d:
                if 'hostname' in d[k].keys():
                    flattened_dict.append(prefix+k)
                else:
                    new_prefix += k+'/'
                    do_flat(d[k], prefix=new_prefix)
                    new_prefix = prefix
            return flattened_dict

    def _flatten_v2(self) -> dict:
        flattened_dict = {}

        def do_flat(d, prefix=''):
            new_prefix = prefix
            for k, v in d.items():
                if isinstance(v, dict):
                    if all(map(lambda i: type(i) == dict, v.values())):
                        # node_type = 'group'
                        new_prefix += k+'/'
                        do_flat(d[k], prefix=new_prefix)
                        new_prefix = prefix
                    elif all(map(lambda i: type(i) != dict, v.values())):
                        # node_type = 'host'
                        flattened_dict[prefix+k] = v
                else:
                    console_logger.error(
                        'Incorrect node type: {}.'.format(k))
                    sys.exit(1)
            return flattened_dict

        return do_flat(self.inventory_dict)

    def get_host(self, hostname: str) -> dict:
        return self.__getitem__(hostname)



import functools
import os

import logging
import logging.config
import sys
import yaml

DEFAULT_CONFIG_DIR = os.path.expanduser('~/.config/pysch')
DEFAULT_CONFIG_FILE = DEFAULT_CONFIG_DIR + '/config.yaml'
DEFAULT_INVENTORY_FILE = DEFAULT_CONFIG_DIR + '/inventory.yaml'
DEFAULT_PWDDB_FILE = DEFAULT_CONFIG_DIR + '/pwddb.kdbx'
DEFAULT_PWDDB_KEY = DEFAULT_CONFIG_DIR + '/pwddb.keyx'

DEFAULT_LOGGING_CONFIG = {
    'version': 1,
    'formatters': {
        'simple': {
            'format': '%(levelname)s - %(message)s'
        },
        'pysc_default': {
            'format': '%(asctime)s [%(levelname)s] %(name)s (%(filename)s).%(funcName)s(%(lineno)d) - %(message)s'
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'simple',
            'stream': 'ext://sys.stdout'
        },
        'log_file': {
            'class': 'logging.FileHandler',
            'level': 'DEBUG',
            'formatter': 'pysc_default',
            'filename': 'pysch.log'
        }
    },
    'loggers': {
        'console_logger': {
            'level': 'DEBUG',
            'handlers': ['console'],
            'propagate': True
        },
        'pysch': {
            'level': 'DEBUG',
            'handlers': ['log_file'],
            'propagate': True
        }
    },
    'root': {
        'level': 'DEBUG',
        'handlers': ['log_file']
    },
    'disable_existing_loggers': False
}


def get_local_terminal_size():
    """"""
    term_size = os.get_terminal_size()
    return term_size.columns, term_size.lines


def get_local_terminal_type():
    return os.getenv('TERM')


def flatten_log_msg(msg):
    return str(msg).replace('\n', ', ')


def singlton_class(cls):
    instance = None

    @functools.wraps(cls)
    def inner(*args, **kwargs):
        nonlocal instance
        if instance is None:
            instance = cls(*args, **kwargs)
        return instance

    return inner


def flat_inventory(inv):
    pass


def configure_logging(loglevel, config_file=DEFAULT_CONFIG_FILE) -> None:

    logger = logging.getLogger(__name__)

    try:
        with open(DEFAULT_CONFIG_FILE, 'r') as f:
            conf_dict = yaml.load(f, yaml.Loader)
    except FileNotFoundError:
        logger.error('File {} not found. Using default logging config...'.format(
            config_file)
        )
        logging_config = DEFAULT_LOGGING_CONFIG
    else:
        try:
            logging_config = conf_dict['logging']
        except KeyError:
            # logger.error('No logging configuration in config. Using default...')
            logging_config = DEFAULT_LOGGING_CONFIG

    try:
        for logger in logging_config['loggers']:
            logging_config['loggers'][logger]['level'] = loglevel
        logging_config['root']['level'] = loglevel
    except KeyError:
        print('Cannot set logging level.',
              'Plese check the loggers configuration.')
        sys.exit(1)

    logging.config.dictConfig(logging_config)



from click.testing import CliRunner
import pykeepass
from pysch import cli, common, credentials
import pytest


class TestCli():

    def test_cli_init(self):
        runner = CliRunner()
        result = runner.invoke(cli.cli, ['init'])
        assert result.exit_code == 0
        assert "Default configuration has been saved at" in result.output

    def test_cli_command_list(self):
        runner = CliRunner()
        result = runner.invoke(cli.cli)
        command_list = [
            'add-credentials',
            'connect',
            'init',
            'list-credentials',
            'list-hosts',
        ]
        assert all([command in result.output for command in command_list])

    def test_cli_connect_no_host(self):
        runner = CliRunner()
        result = runner.invoke(cli.cli, ['connect'])
        assert result.exit_code == 2
        assert "Missing argument 'HOST'" in result.output

    def test_cli_add_credentials(self):
        pass

    def test_cli_add_credentials_duplicate(self):
        pass

    def test_cli_delete_credentials(self):
        pass


class TestCredentialsApi():

    def test_credentials_default_file_open(self):
        cred = credentials.Credentials(
            common.DEFAULT_PWDDB_FILE,
            keyfile=common.DEFAULT_PWDDB_KEY
        )
        assert isinstance(cred, credentials.Credentials)
        assert isinstance(cred._kdbx, pykeepass.PyKeePass)
        assert cred._kdbx.filename == common.DEFAULT_PWDDB_FILE

    def test_credentials_add(self):
        test_entry_name = 'test_pytest'
        cred = credentials.Credentials(
            common.DEFAULT_PWDDB_FILE,
            keyfile=common.DEFAULT_PWDDB_KEY
        )
        entry = cred.add(
            test_entry_name,
            username='test_username',
            password='test_password'
        )
        assert isinstance(entry, pykeepass.entry.Entry)

    def test_credentials_add_duplicate(self):
        test_entry_name = 'test_pytest'
        cred = credentials.Credentials(
            common.DEFAULT_PWDDB_FILE,
            keyfile=common.DEFAULT_PWDDB_KEY
        )
        with pytest.raises(ValueError) as e:
            cred.add(
                test_entry_name,
                username='test_username',
                password='test_password'
            )
        assert "already exists" in str(e.value)

    def test_credentials_get(self):
        test_entry_name = 'test_pytest'
        cred = credentials.Credentials(
            common.DEFAULT_PWDDB_FILE,
            keyfile=common.DEFAULT_PWDDB_KEY
        )
        entry = cred.get(test_entry_name)
        assert entry is not None
        assert isinstance(entry, pykeepass.entry.Entry)
        assert entry.title == test_entry_name

    def test_credentials_get_fail(self):
        cred = credentials.Credentials(
            common.DEFAULT_PWDDB_FILE,
            keyfile=common.DEFAULT_PWDDB_KEY
        )
        entry = cred.get('this entry doesnt exist')
        assert entry is None

    def test_credentials_iter(self):
        cred = credentials.Credentials(
            common.DEFAULT_PWDDB_FILE,
            keyfile=common.DEFAULT_PWDDB_KEY
        )
        entry_list = [item for item in cred]
        all_of_entry_type = all(
            map(
                lambda ent: isinstance(ent, pykeepass.entry.Entry), entry_list
            )
        )
        assert len(entry_list) > 0
        assert all_of_entry_type is True

    def test_credentials_delete(self):
        test_entry_name = 'test_pytest'
        cred = credentials.Credentials(
            common.DEFAULT_PWDDB_FILE,
            keyfile=common.DEFAULT_PWDDB_KEY
        )
        cred.delete(test_entry_name)



import logging
import os
import sys
from random import randbytes
from hashlib import sha256

import click
from lxml import etree
import pykeepass
# from pykeepass.exceptions import CredentialsError
from yaml import dump as yaml_dump
try:
    from yaml import CDumper as YamlDumper
except ImportError:
    from yaml import Dumper as YamlDumper

from .common import (
    configure_logging,
    DEFAULT_CONFIG_DIR,
    DEFAULT_CONFIG_FILE,
    DEFAULT_INVENTORY_FILE,
    DEFAULT_PWDDB_FILE,
    DEFAULT_PWDDB_KEY
)

from .pysch_cli import PyscCLI
from .config import Config
from .credentials import Credentials

console_logger = logging.getLogger('console_logger')
logger = logging.getLogger(__name__)

DEFAULT_CONFIG_FILE_SET = (
    DEFAULT_CONFIG_FILE,
    DEFAULT_INVENTORY_FILE,
    DEFAULT_PWDDB_FILE,
    DEFAULT_PWDDB_KEY
)


@click.group()
@click.option(
    '--config', '-c',
    help='Path to configuration file',
    type=click.Path(exists=False),
    default=DEFAULT_CONFIG_FILE)
@click.option(
    '--loglevel', '-l',
    type=click.Choice(
        ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'],
        case_sensitive=False),
    default='ERROR')
@click.pass_context
def cli(ctx, loglevel, config):

    cfg = os.path.abspath(os.path.expanduser(config)) if config else None
    if config:
        cfg = os.path.abspath(os.path.expanduser(config))
        configure_logging(loglevel, config_file=cfg)
    else:
        cfg = None
        configure_logging(loglevel)
    console_logger.debug('Using config file: {}'.format(config))

    ctx.ensure_object(dict)
    ctx.obj['CONFIG'] = config


@cli.command(help='Create initial config and inventoty')
@click.option(
    '--force',
    is_flag=True,
    help='Rewrite existing configuration files in ~/.config/pysch'
)
def init(force):
    if not os.path.exists(DEFAULT_CONFIG_DIR):
        os.makedirs(DEFAULT_CONFIG_DIR)
    # if os.path.exists(DEFAULT_CONFIG_FILE):
    if any(map(os.path.exists, DEFAULT_CONFIG_FILE_SET)):
        if force:
            for fname in DEFAULT_CONFIG_FILE_SET:
                if os.path.exists(fname):
                    click.echo(
                        'File {} will be overrided!'.format(fname)
                    )
            click.confirm('Do you want to continue?', abort=True)
        else:
            click.echo(
                'Some of configuration files exist in {}'.format(
                    DEFAULT_CONFIG_DIR))
            click.echo('Use --force to override.')
            sys.exit()
    with open(DEFAULT_CONFIG_FILE, 'w') as config_file:
        yaml_dump(
            {
                'inventory_file': DEFAULT_INVENTORY_FILE,
                'keepass_db_file': DEFAULT_PWDDB_FILE,
                'keepass_key_file': DEFAULT_PWDDB_KEY,
            },
            config_file,
            YamlDumper
        )

    with open(DEFAULT_INVENTORY_FILE, 'w') as inventory_file:
        yaml_dump(
            {
                'localhost': {
                    'hostname': '127.0.0.1',
                    'credentials': 'my_credentials',
                },
                'my_group': {
                    'my-host-1': {
                        'hostname': '192.168.1.2',
                        'credentials': 'my_password',
                    },
                    'my-host-2': {
                        'hostname': 'server.local',
                        'credentials': 'my_server_password',
                    },
                },
            },
            inventory_file,
            YamlDumper
        )

        key_bytes = randbytes(32)
        key_hex = key_bytes.hex().upper()
        key_hash = sha256(key_bytes).digest()[:4].hex().upper()
        with open(DEFAULT_PWDDB_KEY, 'wb') as key_file:
            xml_key = etree.XML('<KeyFile><Meta><Version>2.0</Version></Meta>\
                <Key><Data Hash="{}">{}</Data></Key></KeyFile>'.format(
                    key_hash,
                    key_hex
                )
            )
            etree.indent(xml_key)
            xml_tree = etree.ElementTree(xml_key)
            xml_tree.write(
                key_file,
                xml_declaration=True,
                encoding='utf-8',
                pretty_print=True
            )

        pykeepass.create_database(
            DEFAULT_PWDDB_FILE,
            keyfile=DEFAULT_PWDDB_KEY
        )

        click.echo('Default configuration has been saved at {}'.format(
            DEFAULT_CONFIG_DIR
        ))


@cli.command(help='Connect to the host')
@click.argument('host')
@click.option(
    '--session-log',
    help='Session log file location',
    type=click.Path(exists=False))
@click.pass_context
def connect(ctx, host, session_log):
    click.echo('Connecting to {}'.format(host))
    if session_log:
        session_log = os.path.abspath(session_log)
    PyscCLI(config_file=ctx.obj['CONFIG']).connect(
        host,
        session_log=session_log)


@cli.command(help='Get list of hosts')
@click.pass_context
def list_hosts(ctx):
    click.echo('Available hosts:')
    PyscCLI(config_file=ctx.obj['CONFIG']).list_hosts()


@cli.command(help='Get list of credendials')
@click.pass_context
def list_credentials(ctx):
    click.echo('Available credentials:')
    PyscCLI(config_file=ctx.obj['CONFIG']).list_credentials()


@cli.command(help='Add new credendials')
@click.option('--title', '-t', prompt=True)
@click.option('--username', '-u', prompt=True)
@click.option('--password', '-p',
              prompt=True,
              hide_input=True,
              confirmation_prompt=True)
@click.pass_context
def add_credentials(ctx, title, username, password):
    cfg = Config(ctx.obj['CONFIG'])
    pwddb = Credentials(
        cfg.keepass_db_file,
        keyfile=cfg.keepass_key_file
    )
    try:
        pwddb.add(title, username, password)
    except ValueError:
        console_logger.error(
            'Entry "{}" already exists in the root group'.format(title))
        sys.exit(1)
    click.echo('"{}" entry has been added to the keepass db'.format(title))



import logging
import os
import socket
import sys
from typing import List

import paramiko

from .common import (
    get_local_terminal_size,
    get_local_terminal_type,
    DEFAULT_CONFIG_FILE,
)
from .config import Config
from .credentials import Credentials
from .interactive import interactive_shell
from .inventory import Inventory

console_logger = logging.getLogger('console_logger')
logger = logging.getLogger(__name__)


class PyscCLI():

    def __init__(self, config_file=DEFAULT_CONFIG_FILE) -> None:
        self.config = Config(config_file)

        try:
            self._inventory = Inventory(self.config.inventory_file)
            logger.debug('Loaded inventory from file ' +
                         self.config.inventory_file)
        except AttributeError:
            console_logger.error('No inventory file provided!')
            sys.exit(1)

        try:
            self.pwddb = Credentials(
                self.config.keepass_db_file,
                keyfile=self.config.keepass_key_file
            )
        except FileNotFoundError as e:
            console_logger.error(str(e))
            sys.exit(1)
        # except CredentialsError as e:
        #     console_logger.error(str(e)
        #     sys.exit(1)

    @property
    def inventory(self):
        if not hasattr(self, '_inventory'):
            self._inventory = Inventory(self.inventory_file)
        return self._inventory

    def list_hosts(self) -> List:
        for host in self.inventory:
            print(host)

    def list_credentials(self) -> List:
        for entry in self.pwddb:
            print(entry)

    def connect(self, target_host, session_log=None):
        self.target_host = target_host
        connection_config = self.inventory.get_host(self.target_host)
        if not connection_config:
            sys.exit(1)

        credentials = self.pwddb.get(connection_config['credentials'])
        if not credentials:
            console_logger.error(
                'Credentials "{}" not found'.format(
                    connection_config['credentials']
                )
            )
            sys.exit(1)
        try:
            connection_config['username'] = credentials.username
            connection_config['password'] = credentials.password
            connection_config.pop('credentials')
        except AttributeError as e:
            console_logger.error(str(e))
            console_logger.error(
                'Please check credentials "{}"'.format(credentials)
            )
            sys.exit(1)

        client = paramiko.SSHClient()
        client.load_system_host_keys()
        # logger.debug('Host keys:')
        # logger.debug(client._system_host_keys.keys())
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy)
        # logger.info('connecting to '+self.config['hostname'])
        # logger.info('connecting to '+self.config['hostname'])
        # TODO: deal with BadHostKeyException
        # better to check it in advance even
        # client.load_system_host_keys
        # paramiko.HostKeys
        try:
            client.connect(**connection_config)
        except socket.error:
            console_logger.error('Cannot connect to host "{}"'.format(
                connection_config['hostname']
            ))
            sys.exit(1)
        except paramiko.ssh_exception.AuthenticationException:
            console_logger.error('Authentication failed!')
            sys.exit(1)

        t = client.get_transport()
        channel = t.open_session()

        for var_name, var_value in os.environ.items():
            if var_name.startswith('LC_') or var_name == 'LANG':
                channel.set_environment_variable(var_name, var_value)

        try:
            channel.get_pty(
                get_local_terminal_type(),
                *get_local_terminal_size()
            )
            channel.invoke_shell()
        except paramiko.SSHException as e:
            console_logger.error(str(e))
            console_logger.error('Failed to request pty and interactve shell')
            sys.exit(1)

        interactive_shell(channel, session_log)



import pykeepass


class Credentials:

    def __init__(self, filename, password=None, keyfile=None) -> None:
        self._kdbx = pykeepass.PyKeePass(filename, password, keyfile)

    def __len__(self) -> int:
        return len(self._kdbx.entries)

    def __contains__(self, title: str) -> bool:
        if self._kdbx.find_entries(title=title, first=True):
            return True
        return False

    def __iter__(self):
        return iter(self._kdbx.entries)

    def __getitem__(self, title: str) -> pykeepass.entry.Entry:
        item = self._kdbx.find_entries(
            title=title,
            first=True)
        return item

    def get(self, title: str) -> pykeepass.entry.Entry:
        item = self.__getitem__(title)
        return item if item else None

    def add(self, title, username, password) -> pykeepass.entry.Entry:
        try:
            res = self._kdbx.add_entry(
                self._kdbx.root_group, title, username, password)
        except Exception as e:
            res = None
            if 'already exists' in str(e):
                raise ValueError(
                    'Entry "{}" already exists in the root group'.format(
                        title))
            else:
                raise e
        self._kdbx.save()
        return res

    def delete(self, title):
        item = self.__getitem__(title)
        if item:
            item.delete()
            self._kdbx.save()



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name='pysch',
    version='0.1',
    description='',
    py_modules=['pysch'],
    package_dir={'pysch': 'pysch'},
    package_data={},
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'pysch = pysch.cli:cli',
        ],
    },
    install_requires=['paramiko', 'pykeepass', 'pyyaml', 'click'],
)



# Copyright (C) 2003-2007  Robey Pointer <robeypointer@gmail.com>
#
# This file is part of paramiko.
#
# Paramiko is free software; you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# Paramiko is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Paramiko; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.


import fcntl
import logging
import re
import select
import socket
import struct
import sys
import termios
import tty


from .common import get_local_terminal_size

logger = logging.getLogger(__name__)


class SessionLogger():

    csi_escape_seq = re.compile(br'\x1b\[(?:[0-?]+)?(?:[ -\/]+)?[@-~]?')
    # backspace_escape = re.compile(br'\w\x08')
    backspace_escape = re.compile(br'\n(.*\r\r)(?!\n)')
    car_return_escape = re.compile(br'\s+\r')

    def __init__(self, filename) -> None:
        self._buffer = bytearray()
        self.filename = filename
        pass

    def write(self, some_bytes, force=False):
        self._buffer.extend(some_bytes)
        if (len(self._buffer) >= 1024) or force:
            with open(self.filename, 'ab') as f:
                self._buffer = self.csi_escape_seq.sub(b'', self._buffer)
                self._buffer = self.backspace_escape.sub(b'\n', self._buffer)
                self._buffer = self.car_return_escape.sub(b'', self._buffer)
                f.write(self._buffer)
                self._buffer = bytearray()


def interactive_shell(chan, session_log_fname):

    oldtty = termios.tcgetattr(sys.stdin)
    try:
        tty.setraw(sys.stdin.fileno())
        tty.setcbreak(sys.stdin.fileno())
        chan.settimeout(0.0)

        if session_log_fname:
            session_logger = SessionLogger(session_log_fname)

        while True:
            if not chan.closed:
                chan.resize_pty(
                    *get_local_terminal_size()
                )
            r, w, e = select.select([chan, sys.stdin], [], [])
            wat = fcntl.ioctl(r[0].fileno(), termios.FIONREAD, "  ")
            doublewat = struct.unpack('h', wat)[0]
            if chan in r:
                try:

                    x = chan.recv(1024)

                    if len(x) == 0:
                        sys.stdout.buffer.write(b"\r\n*** EOF\r\n")
                        if session_log_fname:
                            session_logger.write(x, force=True)
                        break
                    sys.stdout.buffer.write(x)
                    if session_log_fname:
                        session_logger.write(x)
                    sys.stdout.buffer.flush()
                except socket.timeout:
                    pass
            if sys.stdin in r:
                x = sys.stdin.read(doublewat)
                if len(x) == 0:
                    if session_log_fname:
                        session_logger.write(b'', force=True)
                    break
                chan.send(x)

    finally:
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, oldtty)
